import os.path 
import requests
import urllib

from .decomposition import Decomposition
from .exception import process_errors
from ..auth import Auth
from .sampler import Sampler

API = "https://www.imandra.ai/"

def decompose(code, api_url=None, auth_token=None):
    # Get auth token
    if auth_token == None:
        if decompose.token == None:
            auth = Auth()
            decompose.token = auth.token
        auth_token = decompose.token
            
    headers = { 'X-Auth': auth_token }

    if api_url == None:
        api_url = API

    r = requests.post\
      ( os.path.join(api_url, "api/ipl-worker/enqueue_py")
      , data = code
      , headers = headers
      )
    job_hash = process_errors(r)

    return Decomposition(job_hash, api_url, headers)
   

decompose.token = None




import random
import uuid
import json
import os
import os.path
import urllib

import hashlib
import base64 
import requests

def get_secrets():
    rchar = lambda: random.SystemRandom().choice(get_secrets.chars) 
    verifier = bytes(rchar() for _ in range(32))
    challenge = hashlib.sha256(verifier).digest()
    challenge = base64.b64encode(challenge, altchars=b"-_")
    challenge = challenge.replace(b"=",b"")
    return verifier, challenge

get_secrets.chars = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

class Auth:
    def __init__(self):
        self.auth0_base_host  = "auth.imandra.ai"
        self.auth0_client_id  = "q2yGHBTLmJSia35zCOdkUpEecj9mQl6o"
        self.auth0_audience   = "https://www.imandra.ai/api"
        self.imandra_web_host = "https://www.imandra.ai"
        self.http_port        = None

        self.folder_path      =  os.path.join(os.environ['HOME'], ".imandra")

        self.redirect_uri = self.imandra_web_host + "/pkce-callback"
        if self.http_port:
            self.redirect_uri += "?redirect_port=" + str(self.http_port)

        self.ensure_folder()
        self.ensure_token()
        

    def ensure_folder(self):
        device_id_path = os.path.join(self.folder_path, "device_id")
        if not os.path.exists(self.folder_path):
            os.mkdir(self.folder_path)
        if not os.path.exists(device_id_path):
            self.device_id = str(uuid.uuid4())
            with open(device_id_path, 'w') as device_id_file:
                device_id_file.write(self.device_id)
        else:
            with open(device_id_path, 'r') as device_id_file:
                self.device_id = device_id_file.read()
                
       
    def ensure_token(self):
        token_path = os.path.join(self.folder_path, "login_token")
        if os.path.exists(token_path):
            with open(token_path, 'r') as token_file:
                self.token = token_file.read().strip()
            return
        
        self.verifier, self.challenge = get_secrets()
        link = self.make_link()
        print("We need to authenticate you! Please open: \n{}".format(link))
        code_str = str(input("Enter the code provided on the page above: "))
        self.token = self.exchange_tokens(code_str)
        
        with open(token_path, 'w') as token_file:
            token_file.write(self.token)
        
    def make_link(self):
        params = urllib.parse.urlencode(\
          { "audience"       : self.auth0_audience
          , "response_type"  : "code"
          , "client_id"      : self.auth0_client_id
          , "code_challenge" : self.challenge
          , "code_challenge_method" : "S256"
          , "redirect_uri"   : self.redirect_uri
          })
        return "https://{}/{}?{}".format(self.auth0_base_host, "authorize", params)
    
    
    def exchange_tokens(self, code_str):
        tkn = \
          { "grant_type"    : "authorization_code"
          , "client_id"     : self.auth0_client_id
          , "code_verifier" : self.verifier.decode("utf-8") 
          , "code"          : code_str
          , "redirect_uri"  : self.redirect_uri
          } 
        url = "https://{}/{}".format(self.auth0_base_host, "oauth/token")
        r = requests.post(url, json=tkn)
        if not r.ok:
            raise ValueError(r.text)
        auth_token = r.json()
        headers = \
          { "Authorization" : "Bearer {}".format(auth_token['access_token'])
          , "X-Device-Id"   : self.device_id
          } 
        url = "{}/{}".format(self.imandra_web_host, "api/login-token-for")
        r = requests.get(url, headers=headers)
        if not r.ok:
            raise ValueError(r.text)
        return r.text



import re
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup\
  ( name='imandra'
  , packages=[package for package in find_packages() if package.startswith('imandra')]
  , install_requires=[ "requests" ]  
  , description='Python interface to Imandra (cloud-native automated reasoning engine for analysis of algorithms)'
  , long_description=long_description
  , long_description_content_type="text/markdown"
  , author='Imandra'
  , url='http://imandra.ai'
  , author_email='kostya@aestheticintegration.com'
  , version='0.1.11'
  , classifiers=\
    [ "Programming Language :: Python :: 3"
    , "Operating System :: OS Independent"
    ]

  )



import os.path 
import requests
import json

from .exception import process_errors

class Decomposition(object):
    def __init__(self, job_hash, api_url, headers):
        self.job_hash = job_hash
        self.api_url  = api_url
        self.headers  = headers
        self.__regions_code = \
          { "flat" : None
          , "tree" : None
          }

    def status(self):
        url = os.path.join(self.api_url, "api/ipl-worker/status", self.job_hash)
        r = requests.get\
          ( url
          , headers = self.headers
          )
        return process_errors(r)

    def __get_regions_code(self, kind):
        url = os.path.join(self.api_url, "api/ipl-worker/data", self.job_hash, "regions_ast_Scenario_symbolic_{}.py".format(kind))
        r = requests.get\
          ( url
          , headers = self.headers
          )
        self.__regions_code[kind] = process_errors(r)
 
    def dump(self, fp, kind="tree"):
        if self.__regions_code[kind] is None:
            self.__get_regions_code(kind)
        fp.write(self.__regions_code[kind]) 

    def dumps(self, kind="tree"):
        if self.__regions_code[kind] is None:
            self.__get_regions_code(kind)
        return self.__regions_code[kind]







import json

class PyiplError(Exception):
    def __init__(self, *args, **kwargs):
        self.text = args[0]
        super(PyiplError, self).__init__(*args, **kwargs)
    def __repr__(self):
        return str(self.text)

class ErrorDecodingJSON(PyiplError):
    pass
class ErrorConvertingRegionToPython(PyiplError):
    pass
class PyiplRequestError(PyiplError):
    pass
class ErrorProcessingDeclaredTypes(PyiplError):
    pass
class ErrorProcessingExpressions(PyiplError):
    pass
class ErrorProcessingStateClass(PyiplError):
    pass
class ErrorProcessingActionReceive(PyiplError):
    pass
class IPLError(PyiplError):
    pass
class JobStatusError(PyiplError):
    pass
class GSBucketAccessError(PyiplError):
    pass
class InvalidJSON(PyiplError):
    pass

errdict = \
  { "ErrorDecodingJSON"               : ErrorDecodingJSON
  , "ErrorConvertingRegionToPython"   : ErrorConvertingRegionToPython
  , "PyiplRequestError"               : PyiplRequestError
  , "ErrorProcessingDeclaredTypes"    : ErrorProcessingDeclaredTypes
  , "ErrorProcessingExpressions"      : ErrorProcessingExpressions
  , "ErrorProcessingStateClass"       : ErrorProcessingStateClass
  , "ErrorProcessingActionReceive"    : ErrorProcessingActionReceive
  , "IPLError"                        : IPLError
  , "JobStatusError"                  : JobStatusError
  , "GSBucketAccessError"             : GSBucketAccessError
  }


def process_errors(responce):

    if responce.ok:
        return responce.text

    try:
        err = json.loads(responce.text)
    except:
        raise InvalidJSON("API Server returned invalid JSON: " + responce.text)
   
    if err.get("error") == "ipl-gen":
        raise IPLError(err.get("body"))
            
    if err.get("error") == "py-to-ipl":
        err = err.get("body", {})
        
    etype = err.get("ErrorType", "")
    text = err.get("ErrorBody", "")
    ex = errdict.get(etype, PyiplError)
    raise ex(text)



from collections import defaultdict

class Sampler(object):
    """
    A region sampler class, performs random search, trying to find the least visited region.

    Parameters
    ----------
    calculate_regions : function, required
        Region calculation function, typically, created with imandra.idf.decompose
    random_sequence : function, required
        A random generator of sequences. Ideally, should generate a uniform distribution over
        all allowed input sequences for the model / regions. 

    """
    def __init__(self, calculate_regions, random_sequence):
        self.calculate_regions = calculate_regions
        self.random_sequence = random_sequence
        self.counts = defaultdict(lambda:0, {})
        self.maxCount = 1
        
            
    def generate(self, N, retries=50):
        """
        Runs the input sequence generation, attemting to produce a region-uniform distribution over them.

        Parameters
        ----------
        N : int, required
            Number of many sequences to generate
        retires : int, optional
            Number of rejection retries per an attempt. Defaults to 50. 

        """
        samples = []
        for _ in range(N):
            currentBest, best_query, best_region = self.maxCount, None, None
            for _ in range(retries):
                sequence = self.random_sequence()
                region = tuple(self.calculate_regions(*sequence))
                rcount = self.counts[region]
                if rcount == 0:
                    best_query, best_region = sequence, region
                    break
                if rcount < currentBest:
                    best_query, best_region = sequence, region
                    currentBest = rcount

            if best_query is None:
                best_query, best_region = sequence, region
            self.counts[best_region] += 1
            if self.counts[best_region] > self.maxCount:
                self.maxCount = self.counts[best_region]

            samples.append(best_query)
        return samples

